# More examples are at https://github.com/ibm-messaging/mq-dev-patterns
# and in code/examples in the source distribution.

"""
This example shows use of a dynamic queue as a reply destination. It sends
a message but the code to wait for and read the reply messaage would have to be
added. Alongside working with another application that does the actual reply.
See the reply_to_queues example for one such.

Note that the user must have "get" and "dsp" authorities against the model queue.
"""

import ibmmq as mq

queue_manager = 'QM1'
channel = 'DEV.APP.SVRCONN'
host = '127.0.0.1'
port = '1414'
conn_info = '%s(%s)' % (host, port)
user = 'app'
password = 'password'
message = 'Please reply to a dynamic queue, thanks.'
dynamic_queue_prefix = 'MY.REPLIES.*'
request_queue = 'DEV.QUEUE.1'

qmgr = mq.connect(queue_manager, channel, conn_info, user, password)

# Dynamic queue's object descriptor.
dyn_od = mq.OD()
dyn_od.ObjectName = 'SYSTEM.DEFAULT.MODEL.QUEUE'
dyn_od.DynamicQName = dynamic_queue_prefix

# Open the dynamic queue.
dyn_input_open_options = mq.CMQC.MQOO_INPUT_EXCLUSIVE
dyn_queue = mq.Queue(qmgr, dyn_od, dyn_input_open_options)
dyn_queue_name = dyn_od.ObjectName.strip()

# Prepare a Message Descriptor for the request message.
md = mq.MD()
md.ReplyToQ = dyn_queue_name

# Send the message.
queue = mq.Queue(qmgr, request_queue)
queue.put(message, md)

# Get and process the response here...

dyn_queue.close()
queue.close()
qmgr.disconnect()

print("Done.")



"""Test helpers and utilities"""

import sys
import os
import inspect
import random
import functools  # >= Python 2.5

def with_env_complement(env_var_name, envvar_value):
    """Decorate test with environment additions.

    Use for tests with mandatory environment variable requirements. If env
    variable env_var_name is not set before test invocation, add it to the
    environment with a value of envvar_value, run the test and remove
    env_var_name again.
    Runs the test without environment changes if env_var_name is set.
    """

    # The actual decorator that will accept the test function argument and wrap
    # it with the optional env additions
    def decorate(test_func):
        @functools.wraps(test_func)
        def _env_complemented(*args, **kwargs):
            if env_var_name not in os.environ.keys():  # pylint: disable=C0201
                # enhance environment if env variable is not set
                os.environ[env_var_name] = envvar_value
                try:
                    return test_func(*args, **kwargs)
                finally:
                    del os.environ[env_var_name]
            else:
                # run test as is, without changes to the environment
                return test_func(*args, **kwargs)
        return _env_complemented
    return decorate


# a helper to access the config class attributes
def _visit(cls, prefix=()):
    """Recursively traverse non-special cls attributes and yield
    (qualified-name, value)-tuples. Generator function.

    A "qualified" name is a (prefix, outer-cls, [inner-cls,] attr-name)-tuple.
    Args:
        cls: The (config) class to traverse for attribute lookup.
        prefix: Optional prefix tuple.
    """
    # pylint: disable=R1737
    for attr_name, attr_val in cls.__dict__.items():
        if not attr_name.startswith('__'):
            # ignore names that appear "special"
            if inspect.isclass(attr_val):
                # recursively
                for visited in _visit(
                        attr_val, prefix=prefix + (cls.__name__,)):
                    yield visited
            else:
                yield (prefix + (cls.__name__, attr_name), attr_val)


def print_config(*cfg_classes):
    """Print configuration set in configuration classes cfg_classes.
    """
    print()
    print("Active configuration:")
    for cls in cfg_classes:
        for q_attr_name, attr_val in sorted(_visit(cls, prefix=('config',))):
            print(("  %s: %s" % ('.'.join(q_attr_name), attr_val)).encode(
                sys.stdout.encoding, 'replace'))


def ispy3str(s):
    """returns True if input arg is a python3 string. False otherwise.
    """
    if isinstance(s, str) and not isinstance(s, bytes):
        return True
    return False


def py3str2bytes(s, encoding='ascii'):
    """Sometimes return a string and sometimes return the byte array"""
    if ispy3str(s) and random.randint(0, 1) == 0:
        return s.encode(encoding)
    return s

def strcmp(a, b):
    """Compare two strings that might be either unicode or byte arrays"""
    rc = False
    encoding = 'ascii'

    if a == b:
        rc = True
    if isinstance(a, str) and isinstance(b, bytes):
        if a.encode(encoding) == b:
            rc = True

    if isinstance(b, str) and isinstance(a, bytes):
        if b.encode(encoding) == a:
            rc = True

    return rc



"""MQIIH: IMS Information Header"""

# Copyright (c) 2025 IBM Corporation and other Contributors. All Rights Reserved.
# Copyright (c) 2009-2024 Dariusz Suchojad. All Rights Reserved.

from mqcommon import *
from mqopts import MQOpts
from ibmmq import CMQC

class IIH(MQOpts):
    """ Construct an MQIIH Structure with default values as per MQI.
    The default values may be overridden by the optional keyword arguments 'kw'.

    A message using this structure is expected to then follow it with the IMS transaction
    data: LLZZ<trancode><data>[LLZZ<data>][LLZZ<data>]. This class does not give any
    assistance in building the data elements.
    """
    def __init__(self, **kw):
        opts = [['_StrucId', CMQC.MQIIH_STRUC_ID, '4s'],
                ['Version', CMQC.MQIIH_VERSION_1, MQLONG_TYPE],
                ['StrucLength', CMQC.MQIIH_LENGTH_1, MQLONG_TYPE],
                ['Encoding', 0, MQLONG_TYPE],
                ['CodedCharSetId', 0, MQLONG_TYPE],
                ['Format', CMQC.MQFMT_NONE, '8s'],
                ['Flags', CMQC.MQIIH_NONE, MQLONG_TYPE],
                ['LTermOverride', ' ', '8s'],
                ['MFSMapName', ' ', '8s'],
                ['ReplyToFormat', CMQC.MQFMT_NONE, '8s'],
                ['Authenticator', CMQC.MQIAUT_NONE, '8s'],
                ['TranInstanceId', CMQC.MQITII_NONE, '16s'],
                ['TranState', CMQC.MQITS_NOT_IN_CONVERSATION, 'b'],
                ['CommitMode', CMQC.MQICM_COMMIT_THEN_SEND, 'b'],
                ['SecurityScope', CMQC.MQISS_CHECK, 'b'],
                ['_Reserved', 0, 'b']]

        super().__init__(tuple(opts), **kw)



"""MQSCO: SSL/TLS Control Options"""

# Copyright (c) 2025 IBM Corporation and other Contributors. All Rights Reserved.
# Copyright (c) 2009-2024 Dariusz Suchojad. All Rights Reserved.

from mqcommon import *
from mqopts import MQOpts
from ibmmq import CMQC, ibmmqc

# SCO Class for SSL Support courtesy of Brian Vicente (mailto:sailbv@netscape.net)
class SCO(MQOpts):
    """ Construct an MQSCO Structure with default values as per MQI.
    The default values may be overridden by the optional keyword arguments 'kw'.
    The default structure is chosen to be VERSION_5, which was MQ 8.0
    """
    def __init__(self, **kw):
        sco_current_version = ibmmqc.__strucversions__.get("sco", 1)

        opts = [['_StrucId', CMQC.MQSCO_STRUC_ID, '4s'],
                ['Version', CMQC.MQSCO_VERSION_5, MQLONG_TYPE],
                ['KeyRepository', b'', '256s'],
                ['CryptoHardware', b'', '256s'],
                ['AuthInfoRecCount', (0), MQLONG_TYPE],
                ['AuthInfoRecOffset', (0), MQLONG_TYPE],
                ['AuthInfoRecPtr', 0, 'P'],
                ['KeyResetCount', (0), MQLONG_TYPE],
                ['FipsRequired', (0), MQLONG_TYPE],
                ['EncryptionPolicySuiteB', [1, 0, 0, 0], '4' + MQLONG_TYPE],
                ['CertificateValPolicy', (0), MQLONG_TYPE],
                ['CertificateLabel', b'', '64s']]

        if sco_current_version >= CMQC.MQSCO_VERSION_6:
            opts += [['KeyRepoPassword', 0, 'P'],
                     ['_KeyRepoPasswordOffset', 0, MQLONG_TYPE],
                     ['_KeyRepoPasswordLength', 0, MQLONG_TYPE]]

        if sco_current_version >= CMQC.MQSCO_VERSION_7:
            opts += [['HTTPSCertValidation', 0, MQLONG_TYPE],
                     ['HTTPSCertRevocation', 0, MQLONG_TYPE],
                     ['HTTPSKeyStore', 0, 'P'],
                     ['_HTTPSKeyStoreOffset', 0, MQLONG_TYPE],
                     ['_HTTPSKeyStoreLength', 0, MQLONG_TYPE]]

        super().__init__(tuple(opts), **kw)


# Backward compatibility
sco = SCO  # pylint: disable=invalid-name



"""Topic class: for MQOPEN, MQSUB, MQPUT (ie publish), MQCLOSE
"""

# Copyright (c) 2025,2026 IBM Corporation and other Contributors. All Rights Reserved.
# Copyright (c) 2009-2024 Dariusz Suchojad. All Rights Reserved.

from mqcommon import *
from mqerrors import *
from ibmmq import CMQC, OD, PMO, ibmmqc, MQObject
from mqqargs import common_q_args
from mqsub import *

from mqqmgr import *

import mqlog

# Publish Subscribe support - Hannes Wagener 2011
class Topic(MQObject):
    """ Topic encapsulates all the Topic I/O operations, including
    publish/subscribe.  A QueueManager object must be already
    connected. The Topic may be opened implicitly on construction, or
    the open may be deferred until a call to open(), pub() or sub()
    The same as for Queue.

    The Topic to open is identified by a topic name and/or a topic
    string (in which case a default MQOD structure is created using
    those names), or by passing a ready constructed MQOD class.

    Refer to the 'Using topic strings' section in the MQ documentation
    for an explanation of how the topic name and topic string are combined
    to identify a particular topic.
    """

    def __real_open(self):
        """ Really open the topic.  Only do this in pub()?
        """
        mqlog.trace_entry("topic:__real_open")
        if self.__topic_desc is None:
            mqlog.trace_exit("topic:__real_open", ep=1)
            raise PYIFError('The Topic Descriptor has not been set.')

        rv = ibmmqc.MQOPEN(self.__queue_manager.getHandle(),
                           self.__topic_desc.pack(), self.__open_opts)
        if rv[-2]:
            mqlog.trace_exit("topic:__real_open", ep=2, rc=rv[-1])
            raise MQMIError(rv[-2], rv[-1])

        _ = self.__topic_handle = rv[0]
        _ = self.__topic_desc.unpack(rv[1])
        mqlog.trace_exit("topic:__real_open")

    def __init__(self, queue_manager, topic_name=None, topic_string=None, topic_desc=None, open_opts=None):
        """ Associate a Topic instance with the QueueManager object 'queue_manager'
        and optionally open the Topic.

        If topic_desc is passed ignore topic_string and topic_name.

        If open_opts is passed, it specifies topic open options, and
        the topic is opened immediately. If open_opts is not passed,
        the queue open is deferred to a subsequent call to open(),
        pub().

        The following table clarifies when the Topic is opened:

        topic_desc  open_opts   When opened
             N       N       open()
             Y       N       open() or pub()
             Y       Y       Immediately
        """
        mqlog.trace_entry("topic:__init__")

        queue_manager = ensure_strings_are_bytes(queue_manager)
        topic_name = ensure_strings_are_bytes(topic_name)
        topic_string = ensure_strings_are_bytes(topic_string)

        self.__queue_manager = queue_manager
        self.__topic_handle = None
        self.__topic_desc = topic_desc
        self.__open_opts = open_opts

        self.topic_name = topic_name
        self.topic_string = topic_string

        if self.__topic_desc:
            if self.__topic_desc['ObjectType'] is not CMQC.MQOT_TOPIC:
                mqlog.trace_exit("topic:__init__", ep=1)
                raise PYIFError('The Topic Descriptor ObjectType is not MQOT_TOPIC.')
            if self.__topic_desc['Version'] < CMQC.MQOD_VERSION_4:
                mqlog.trace_exit("topic:__init__", ep=2)
                raise PYIFError('The Topic Descriptor Version must be at least MQOD_VERSION_4.')
        else:
            self.__topic_desc = self.__create_topic_desc(topic_name, topic_string)

        if self.__open_opts:
            self.__real_open()
        object_name = None
        if topic_name:
            object_name = topic_name
        if topic_string:
            if object_name:
                object_name = object_name + "/" + topic_string
            else:
                object_name = topic_string
        super().__init__(object_name)
        mqlog.trace_exit("topic:__init__")

    @staticmethod
    def __create_topic_desc(topic_name, topic_string):
        """ Creates a topic object descriptor from a given topic_name/topic_string.
        """
        topic_name = ensure_strings_are_bytes(topic_name)
        topic_string = ensure_strings_are_bytes(topic_string)

        topic_desc = OD()
        topic_desc['ObjectType'] = CMQC.MQOT_TOPIC
        topic_desc['Version'] = CMQC.MQOD_VERSION_4

        if topic_name:
            topic_desc['ObjectName'] = topic_name

        if topic_string:
            topic_desc.set_vs('ObjectString', topic_string, 0, 0, 0)

        return topic_desc

    def __del__(self):
        """ Close the Topic, if it has been opened.
        """
        mqlog.trace_entry("topic:__del__")

        try:
            if self.__topic_handle:
                self.close()
        except (PYIFError, MQMIError):
            pass
        mqlog.trace_exit("topic:__del__")

    def open(self, topic_name=None, topic_string=None, topic_desc=None, open_opts=None):
        """ Open the Topic specified by topic_desc or create a object descriptor
        from topic_name and topic_string.
        If open_opts is passed, it defines the
        Topic open options, and the Topic is opened immediately. If
        open_opts is not passed, the Topic open is deferred until a
        subsequent pub() call.
        """
        mqlog.trace_entry("topic:open")
        topic_name = ensure_strings_are_bytes(topic_name)
        topic_string = ensure_strings_are_bytes(topic_string)

        if self.__topic_handle:
            mqlog.trace_exit("topic:open", ep=1)
            raise PYIFError('The Topic is already open.')

        if topic_name:
            self.topic_name = topic_name

        if topic_string:
            self.topic_string = topic_string

        if topic_desc:
            self.__topic_desc = topic_desc
        else:
            self.__topic_desc = self.__create_topic_desc(self.topic_name, self.topic_string)

        if open_opts:
            self.__open_opts = open_opts
            self.__real_open()
        mqlog.trace_exit("topic:open")

    def pub(self, msg, *opts):
        """ Publish the string buffer 'msg' to the Topic. If the Topic is not
        already open, it is opened now. with the option 'MQOO_OUTPUT'.

        msg_desc is the MD() Message Descriptor for the
        message. If it is not passed, or is None, then a default md()
        object is used.

        put_opts is the PMO() Put Message Options structure
        for the put call. If it is not passed, or is None, then a
        default pmo() object is used.

        If msg_desc and/or put_opts arguments were supplied, they may be
        updated by the put operation.
        """
        mqlog.trace_entry("topic:pub")

        msg_desc, put_opts = common_q_args(*opts)

        if not isinstance(msg, bytes):
            if isinstance(msg, str):
                msg = msg.encode(self.__queue_manager.bytes_encoding)
                msg_desc.CodedCharSetId = self.__queue_manager.default_ccsid
                msg_desc.Format = CMQC.MQFMT_STRING
            else:
                error_message = 'Message type is {0}. Convert to bytes.'
                mqlog.trace_exit("topic:pub", ep=1)
                raise TypeError(error_message.format(type(msg)))

        if put_opts is None:
            put_opts = PMO()

        # If queue open was deferred, open it for put now
        if not self.__topic_handle:
            self.__open_opts = CMQC.MQOO_OUTPUT
            self.__real_open()

        if OTelFunctions.put_trace_before:
            OTelFunctions.put_trace_before(self.__queue_manager, msg_desc, put_opts, msg)

        # Now send the message
        rv = ibmmqc.MQPUT(self.__queue_manager.getHandle(), self.__topic_handle, msg_desc.pack(), put_opts.pack(), msg)
        if rv[-2]:
            mqlog.trace_exit("topic:pub", ep=2, rc=rv[-1])
            raise MQMIError(rv[-2], rv[-1])

        _ = msg_desc.unpack(rv[0])
        _ = put_opts.unpack(rv[1])

        if OTelFunctions.put_trace_after:
            OTelFunctions.put_trace_after(self.__queue_manager, put_opts)
        mqlog.trace_exit("topic:pub")

    # Create an alias as the underlying MQ verb is MQPUT
    put = pub

    def pub_rfh2(self, msg, *opts):
        """pub_rfh2(msg[, msgDesc, putOpts, [rfh2_header, ]])
        Put a RFH2 message. opts[2] is a list of RFH2 headers.
        MQMD and RFH2's must be correct.
        """
        mqlog.trace_entry("topic:pub_rfh2")

        rfh2_buff = b''
        if len(opts) >= 3:
            if opts[2] is not None:
                if not isinstance(opts[2], list):
                    mqlog.trace_exit("topic:pub_rfh2", ep=1)
                    raise TypeError('Third item of opts should be a list.')
                encoding = CMQC.MQENC_NATIVE
                if opts[0] is not None:
                    mqmd = opts[0]
                    encoding = mqmd['Encoding']

                for rfh2_header in opts[2]:
                    if rfh2_header is not None:
                        rfh2_buff = rfh2_buff + rfh2_header.pack(encoding)
                        encoding = rfh2_header['Encoding']

                msg = rfh2_buff + ensure_strings_are_bytes(msg)
            self.pub(msg, *opts[0:2])
        else:
            self.pub(msg, *opts)
        mqlog.trace_exit("topic:pub_rfh2")

    # Create an alias as the underlying MQ verb is MQPUT
    put_rfh2 = pub_rfh2

    def sub(self, *opts):
        """ Subscribe to the topic and return a Subscription object.
        A subscription to a topic can be made using an existing queue, either
        by pasing a Queue object or a string at which case the queue will
        be opened with default options.
        """
        mqlog.trace_entry("topic:sub")

        sub_desc = None
        if len(opts) > 0:
            sub_desc = opts[0]

        sub_queue = None
        if len(opts) > 1:
            sub_queue = ensure_strings_are_bytes(opts[1])

        sub = Subscription(self.__queue_manager)
        sub.sub(sub_desc=sub_desc, sub_queue=sub_queue, topic_name=self.topic_name, topic_string=self.topic_string)

        mqlog.trace_exit("topic:sub")
        return sub

    def close(self, options=CMQC.MQCO_NONE):
        """ Close the topic, using options.
        """
        mqlog.trace_entry("topic:close")

        if not self.__topic_handle:
            raise PYIFError('Topic not open.')

        rv = ibmmqc.MQCLOSE(self.__queue_manager.getHandle(), self.__topic_handle, options)
        if rv[0]:
            mqlog.trace_exit("topic:close", ep=1, rc=rv[-1])
            raise MQMIError(rv[-2], rv[-1])

        self.__topic_handle = None
        self.__topic_desc = None
        self.__open_opts = None
        mqlog.trace_exit("topic:close")
